/*
 *  Copyright (C) 2002-25 - ntop.org
 *
 *  http://www.ntop.org/
 *
 * This code is proprietary code subject to the terms and conditions
 * defined in LICENSE file which is part of this source code package.
 */

#ifndef _BUCKET_H_
#define _BUCKET_H_

/* ********************************** */

#define MAGIC_NUMBER   67

/* ********************************** */

/*
 * fallbacks for essential typedefs
 */
#ifdef WIN32
#ifndef __GNUC__
typedef unsigned char  u_char;
typedef unsigned short u_int16_t;
typedef unsigned int   uint;
typedef unsigned long  u_long;
#endif
typedef u_char  u_int8_t;
typedef u_int16_t u_int16_t;
typedef uint   u_int32_t;
#endif /* WIN32 */

/* ********************************** */

#define NPROBE_FD_SET(n, p)   (*(p) |= (1 << (n)))
#define NPROBE_FD_CLR(n, p)   (*(p) &= ~(1 << (n)))
#define NPROBE_FD_ISSET(n, p) (*(p) & (1 << (n)))
#define NPROBE_FD_ZERO(p)     (*(p) = 0)


#define MAX_PAYLOAD_LEN          1400 /* bytes */

#define FLAG_NW_LATENCY_COMPUTED           1
#define FLAG_APPL_LATENCY_COMPUTED         2
#define FLAG_FRAGMENTED_PACKET_SRC2DST     3
#define FLAG_FRAGMENTED_PACKET_DST2SRC     4

#define MAX_NUM_BYTES_ENTROPY_PKTS        64
#define MAX_NUM_IAT_ENTROPY_PKTS          11
#define NPROBE_UNKNOWN_VALUE               0
#define NPROBE_UNKNOWN_VALUE_STR          "0"

#define nwLatencyComputed(a)          (a && NPROBE_FD_ISSET(FLAG_NW_LATENCY_COMPUTED,   &(a->flags)))
#define applLatencyComputed(a)        (a && NPROBE_FD_ISSET(FLAG_APPL_LATENCY_COMPUTED, &(a->flags)))


#ifdef WIN32

#define _WS2TCPIP_H_ /* Avoid compilation problems */
#define HAVE_SIN6_LEN

/* IPv6 address */
/* Already defined in WS2tcpip.h */
struct win_in6_addr
{
  union
  {
    u_int8_t u6_addr8[16];
    u_int16_t u6_addr16[8];
    u_int32_t u6_addr32[4];
  } in6_u;
#ifdef s6_addr
#undef s6_addr
#endif

#ifdef s6_addr16
#undef s6_addr16
#endif

#ifdef s6_addr32
#undef s6_addr32
#endif

#define s6_addr                 in6_u.u6_addr8
#define s6_addr16               in6_u.u6_addr16
#define s6_addr32               in6_u.u6_addr32

};

#define in6_addr win_in6_addr

struct ip6_hdr {
  union {
    struct ip6_hdrctl {
      u_int32_t ip6_un1_flow;   /* 4 bits version, 8 bits TC,
				   20 bits flow-ID */
      u_int16_t ip6_un1_plen;   /* payload length */
      u_int8_t  ip6_un1_nxt;    /* next header */
      u_int8_t  ip6_un1_hlim;   /* hop limit */
    } ip6_un1;

    u_int8_t ip6_un2_vfc;       /* 4 bits version, top 4 bits tclass */
  } ip6_ctlun;

  struct in6_addr ip6_src;      /* source address */
  struct in6_addr ip6_dst;      /* destination address */
};

/* Generic extension header.  */
struct ip6_ext {
  u_int8_t  ip6e_nxt;		/* next header.  */
  u_int8_t  ip6e_len;		/* length in units of 8 octets.  */
};

#else /* WIN32 */

#ifndef s6_addr32
#ifdef __linux__
#define s6_addr32 in6_u.u6_addr32
#else
#if defined(sun)
#define	s6_addr32	_S6_un._S6_u32
#else
#define s6_addr32 __u6_addr.__u6_addr32
#endif
#endif
#endif
#endif /* WIN32*/

/* ********************************** */

#define MAX_NUM_MPLS_LABELS     10
#define MPLS_LABEL_LEN           3

/* ********************************** */

/*
   NOTE

   whenever you change this datastructure
   please update sortFlowIndex()
*/
typedef struct flow_index {
  u_int32_t srcHost, dstHost;
  u_int16_t sport, dport, subflow_id, vlanId;
  u_int8_t proto, tos, _pad, _pad1;
} FlowIndex;

/* ********************************** */

typedef struct ipAddress {
  u_int8_t ipVersion:3 /* Either 4 or 6 */,
    localHost:1, /* -L: filled up during export not before (see exportBucket()) */
    localHostChecked:1,
    notUsed:3 /* Future use */;

  union {
    struct in6_addr ipv6;
    u_int32_t ipv4; /* Host byte code */
  } ipType;
} IpAddress;

struct mpls_labels {
  u_int16_t numMplsLabels;
  u_char mplsLabels[MAX_NUM_MPLS_LABELS][MPLS_LABEL_LEN];
};

struct pluginEntryPoint; /* engine.h */

typedef struct pluginInformation {
  struct pluginEntryPoint *pluginPtr;
  void *pluginData;
  u_int8_t plugin_used;
  struct pluginInformation *next;
} PluginInformation;

/*
 * If the host is local then stats points to a valid
 * memory area, otherwise it points to NULL
 */

#if defined(HAVE_MAXMINDDB)
/* Emulate a Geo structure that holds
 latitude, longitude, country code and city.
 We don't define all the fields as they are not used

See: https://github.com/maxmind/geoip-api-c/blob/204cc59067839c06cc0f7d27e23c06b1c2555ead/libGeoIP/GeoIPCity.h#L32 */
typedef struct {
  char *country_code;
  char *city;
  float latitude;
  float longitude;
} GeoIPRecord;
#endif

typedef struct hostInfo {
  u_char macAddress[6];
#ifdef HAVE_NETFILTER
  u_int8_t dstMacSet; /* used with netfilter */
#endif
  u_int8_t mask;
  u_int16_t ifIdx;
  u_int32_t asn;
  u_int8_t addr_version; /* Either 4 or 6 */
  union {
    u_int32_t ifHost4;
    struct in6_addr ifHost6;
  } addr;
#if defined(HAVE_MAXMINDDB)
  GeoIPRecord *geo;
  char *asname;
#endif
  char *collected_country_code, *collected_city;
  u_int8_t aspath_len; /* Number of entries != 0 in aspath */
  u_int32_t *aspath; /* If allocated it will be MAX_AS_PATH_LEN long */
} HostInfo;

/* *************************************** */

typedef enum {
  direction_new_flow = 0,
  direction_continuation_flow_same_as_original, /* 1 */
  direction_continuation_flow_opposite_of_original  /* 2 */
} OriginalFlowDirection;

/* *************************************** */

typedef enum {
  flow_message = 0,
  event_message,
  counter_message,
  flow_template_message,
  option_template_values,
  snmp_interfaces_message,
  hello_message, /* Startup hello message */
  listening_ports_message, /* List of open ports */
  custom_ie_message, /* --load-custom-fields */

  /*
     NOTE:
     Leave as last entry and make sure
     it's less that MAX_NUM_KAFKA_TOPICS
  */
  last_message_topic_type
} MessageTopicType;

/* *************************************** */

/* See https://www.iana.org/assignments/ipfix/ipfix.xhtml */
typedef enum {
  flow_end_reason_unknown              = 0x00,
  flow_end_reason_idle_timeout         = 0x01,
  flow_end_reason_active_timeout       = 0x02,
  flow_end_reason_end_of_flow_detected = 0x03,
  flow_end_reason_forced_end           = 0x04,
  flow_end_reason_lack_of_resources    = 0x05,
} FlowEndReason;

/* *************************************** */

typedef struct tv {
  u_int32_t tv_sec, tv_usec;
} _tv;

typedef struct {
  u_int32_t num_pkts_up_to_128_bytes, num_pkts_128_to_256_bytes,
    num_pkts_256_to_512_bytes, num_pkts_512_to_1024_bytes,
    num_pkts_1024_to_1514_bytes, num_pkts_over_1514_bytes;
  u_int16_t smallest_ip_pkt_len, largest_ip_pkt_len;
} IPStats;

typedef struct {
  u_int32_t num_pkts_eq_1, num_pkts_2_5,
  num_pkts_5_32, num_pkts_32_64, num_pkts_64_96,
    num_pkts_96_128, num_pkts_128_160, num_pkts_160_192,
    num_pkts_192_224, num_pkts_224_255;
} TTLStats;

struct tcp_seq_num {
  u_int16_t last_ip_id;
  u_int32_t last, next;
  u_int32_t last_ack, last_segment_len;
  u_int16_t last_window_size;
  u_int8_t last_flags;
};

#define MAX_NUM_SPLT_POOL_PKTS    0xFF /* Must be < 256 as SPLT counters are u_int8_t */
//#define SPLT_NUM_BINS             6
#define SPLT_NUM_BINS              47 /* 32 * 47 = 1504 */

// #define MARKOV_TEST

struct splt_host_stats {
  /* https://www.toptal.com/algorithms/shazam-it-music-processing-fingerprinting-and-recognition */
  /* NOTE: packets are fingerprinted as they arrive */
  FlowDirection last_direction;
  u_int8_t num_fingerprinted_pkts;
  struct timeval c2s_last_packet_ts, s2c_last_packet_ts;

  /* 64, 128, 256, 512, 1024, 1518+ [6 classes] */
  struct ndpi_bin src2dst_plen_distribution, dst2src_plen_distribution;

  /* < 1, < 5, < 10, < 50, < 100, over [6 classes] */
  struct ndpi_bin src2dst_timediff_ms, dst2src_timediff_ms;

  int16_t plen[MAX_NUM_SPLT_POOL_PKTS]; /* +=c2s, -=s2c */

#ifdef MARKOV_TEST
  /* Markov Chains */
  u_int8_t last_plen_bin, plen_markov_chain[SPLT_NUM_BINS*SPLT_NUM_BINS];
  u_int8_t last_timediff_bin, timediff_markov_chain[SPLT_NUM_BINS*SPLT_NUM_BINS];
#endif
};

#define MAX_NUM_ENTROPY_PKTS 16

struct entropy_stats_direction {
  u_int8_t num_pkts;

  float bytes_entropy;
  u_int32_t bytes_distribution[256];

#if 0
  float iat_entropy;
  /* < 10 ms, < 20 ms .. < 90 ms, 100ms+ [11 slots] */
  u_int32_t iat_distribution[11];
#endif
};

struct payload_entropy_stats {
  struct entropy_stats_direction c2s, s2c;
};

typedef struct {
  struct mpls_labels *mplsInfo;

  struct {
    /* This entry is filled only in case of tunneled addresses */
    u_int8_t proto;
    IpAddress src, dst;
    u_int16_t sport, dport;
  } untunneled;

  struct {
    char *ssap, *dsap;
  } osi;

  struct {
    char *ssid;
    u_int8_t access_point_mac[6];
  } wifi;

  struct timeval synTime, synAckTime; /* network Latency (3-way handshake) */

  struct {
    u_int32_t src2dst_max, src2dst_min, src2dst_bytes, src2dst_alreadyset,
      dst2src_max, dst2src_min, dst2src_bytes, dst2src_alreadyset,
      src2dst_avg_bytes, src2dst_avg_ms,
      dst2src_avg_bytes, dst2src_avg_ms;
    struct timeval src2dst_begin, src2dst_last, dst2src_begin, dst2src_last;
  } throughput;

  struct {
    IPStats src2dst, dst2src;
  } ipstats;

  /* TCP Sequence number counters */
  struct {
    struct tcp_seq_num src2dst, dst2src;
  } tcpseq;

  struct {
    TTLStats src2dst, dst2src;
  } ttlstats;

  struct splt_host_stats *splt;
  struct payload_entropy_stats *entropy;

  /*
    client <--------------> nprobe <----------------------> server
    |<- clientNwLatency ->|        |<- serverNwLatency --------->|
    |<--------------- network delay/latency -------------------->|
  */
  struct timeval clientNwLatency; /* The RTT/2 between the client and nprobe */
  struct timeval serverNwLatency; /* The RTT/2 between nprobe and the server */
  struct timeval src2dstApplLatency, dst2srcApplLatency; /* Application Latency */

  union {
    struct {
      u_int32_t last_cli_ack, last_srv_ack;
      struct timeval last_cli_ts, last_srv_ts;
      struct ndpi_analyze_struct cli_to_srv, srv_to_cli;
    } tcp;

    struct {
      struct timeval first_cli_to_srv, first_srv_to_cli, second_cli_to_srv; /* Time of the first packet in each direction */
      float clientRTT3WH, serverRTT3WH; /* Computed at 3WH (msec) */
      bool last_spin_set;
      struct timeval last_ts;
      struct ndpi_analyze_struct cli_min_rtt /* cli <-> ntopng RTT */, srv_min_rtt /* ntopng <-> dst RTT */;
    } udp;
  } rtt; /* Computed continuously */

  struct {
    u_int8_t src_to_dst, dst_to_src;
  } qoe;

  /* Host label */
  struct {
    /*
       NOTE:
       the label strings are pointers to readOnlyGlobals.hostLabels
       so they cannot be freed
    */
    char *src_host, *dst_host;
  } label;
} FlowHashBucketExtensions;

/* *************************************** */

struct flowHashBucket; /* Forward */

typedef struct {
  struct flowHashBucket *prev, *next;
} CircularList;

/* *************************************** */

typedef struct {
  IpAddress src, dst;
  u_int16_t sport, dport;
  u_int8_t proto; /* protocol (e.g. UDP/TCP..) */
} IPKey;

typedef struct {
  u_int32_t teid;
  u_int16_t cell_id;
} GTPKey;

typedef struct {
  u_int8_t src[6], dst[6];
} MacKey;

/* *************************************** */

typedef struct flowHashBucketKeyFields {
  u_int8_t is_gtp_flow:1, is_ip_flow:7; /* 1=IPv4/v6 flow, 0=Ethernet (no IP) flow */
  u_int16_t vlanId;

  union {
    IPKey ipKey;
    MacKey macKey;
    GTPKey gtpKey;
  } k;

} FlowHashBucketKeyFields;

/* *************************************** */

typedef struct flowHashBucketCoreFields {
  u_int32_t flow_idx, flow_hash, flow_serial;
  u_int8_t do_not_expire_for_max_duration; /* Flags */

  /* Key */
  FlowHashBucketKeyFields key;

  /* Value */
  struct {
    struct timeval firstSeenSent, lastSeenSent;
    struct timeval firstSeenRcvd, lastSeenRcvd;
  } flowTimers;

  struct {
    u_int64_t pktSent, pktRcvd;
    u_int64_t bytesSent, bytesRcvd;
  } flowCounters;

} FlowHashBucketCoreFields;

/* *************************************** */

#define FLOW_MARKER_NOTYET 0
#define FLOW_MARKER_PASS   1
#define FLOW_MARKER_DROP   2
/* Custom marker 3..255 */

typedef u_int8_t flow_marker;

#define NTOP_QOE_MAX        100
#define NTOP_QOE_UNKNOWN    255
#define NTOP_QOE_EXCELLENT   90
#define NTOP_QOE_GOOD        75
#define NTOP_QOE_FAIR        65
#define NTOP_QOE_DEGRADED    50
#define NTOP_QOE_POOR        30

/* *************************************** */

#define NO_PROTO_TYPE           0
#define NDPI_PROTO_TYPE         1

typedef struct flowHashMicroBucket {
  FlowHashBucketCoreFields tuple; /* Flow core fields */
  u_int8_t dont_export_flow, /*
			       Set it to 1 if when the flow has to be exported, its memory
			       will be freed but the flow will not be exported
			     */
	  absoluteFlowCounters;
  u_int8_t engine_type, engine_id; /* 0=use default */

  OriginalFlowDirection originalDirection;
  FlowEndReason flow_end_reason;

  struct direction {
    u_int8_t src2dst, dst2src; /* 1=RX [receive], 0=TX [transmit] (packet direction) */
  } rx_direction;

  /* L7 protocol */
  struct {
    u_int8_t firewall_event;

    struct {
      ndpi_confidence_t confidence;
      u_int8_t detection_completed:1, ndpi_giveup_called:1, encrypted_flow:1, _pad:5;
      ndpi_protocol ndpi_proto;
      struct ndpi_flow_struct *flow;
    } ndpi;
  } l7;

  /* Flow -> User Mapping */
  struct {
    u_int8_t user_searched; /* 0=we have not yet tried to match the user
			       1=we have already tried to match the user
			       thus if username==NULL it means that
			       we failed
			    */
    char *username, *gtp_gw1, *gtp_gw2, *apn;
  } user;

  flow_marker verdict;
  u_int8_t bucket_expired; /* Force bucket to expire */
  u_int8_t purge_at_next_loop;

  CircularList hash; /* Hash collision list pointers */

  /* Expire List (max flow duration) */
  CircularList max_duration;

  /* Idle flows (no traffic [idle]) */
  CircularList no_traffic;
} FlowHashMicroBucket;

/* *************************************** */

typedef struct {
  u_int16_t tcpMinWindow, tcpMaxWindow, tcpLastWindow, tcpMaxSegmentSize;
  u_int8_t tcpWinScale;
} TcpWindowStats;

typedef struct {
  IpAddress srcaddr;
  IpAddress dstaddr;
  u_int16_t srcTransportPort;
  u_int16_t dstTransportPort;
  u_int8_t realm;
  u_int8_t event;
} natFields;

typedef struct {
  struct timeval last_pkt_time;
  struct ndpi_analyze_struct *iat;
} iat_stats;

typedef enum {
  packet_to_flow = 0,          /* packet capture or nTap  */
  collected_netflow_ipfix = 1, /* collected NetFlow/IPFIX */
  collected_sflow = 2          /* collected sFlow/nfLite  */
} flow_source;

typedef struct {
  u_int8_t thread_id;      /* Thread on which the bucket was allocated */
  u_int32_t subflow_id;    /*
			     Usually is 0: user for subflows on UDP-based proto such as DNS
			     or sequence number in GTP
			   */
  u_int8_t sampled_export_flow:1,   /* 0=normal flow, 1=sampled flow (i.e. to discard) */
    blacklisted_flow:1, swapped_flow:1, __notused:4;
  flow_source source_nature;
  u_int32_t src2dst_tunnel_id, dst2src_tunnel_id;   /* E.g. GTP tunnel */
  u_int16_t src2dst_session_id, dst2src_session_id; /* E.g. L2TP sessionId */

  u_int32_t if_input, if_output;
  u_int8_t src2dstTos, dst2srcTos;
  u_int8_t src2dstMinTTL, dst2srcMinTTL, src2dstMaxTTL, dst2srcMaxTTL;
  IpAddress nextHop, bgpNextHop;
  HostInfo srcInfo, dstInfo; /* src and dst host metadata information */
  u_int16_t innerVlanId, outerVlanId, vlanId_dst;
  u_int32_t src_peer_as, dst_peer_as, ingress_vrfid, egress_vrfid;
  u_int16_t nat_port_range_start, nat_port_range_end;
  u_int32_t nbar2_application_id;
  FlowHashBucketExtensions *extensions;
  char *collected_interface_name, *collected_account_id;

  struct {
    u_int64_t imsi;
    u_int32_t src_addr;
    u_int8_t postnat_updated;
  } nokia;

  struct {
    iat_stats src_to_dst, dst_to_src;
  } iat;

  /* **************** */

  struct {
    u_int16_t sentFragPkts, rcvdFragPkts;

    struct {
      u_int16_t longest, shortest;
    } pktSize; /* bytes */
  } flowCounters;

  struct {
    u_int16_t lastIdx;
    u_int32_t initialTime;
    u_int32_t *src2dst, *dst2src;
  } secondByteCounters;

  union {
    struct {
      u_int32_t bytesRcvdRetransmitted, pktRcvdRetransmitted;
      u_int32_t bytesSentRetransmitted, pktSentRetransmitted;
      u_int32_t sentOOOrder, rcvdOOOrder;
      u_int32_t sentKeepAlive, rcvdKeepAlive;
      u_int16_t src2dstTcpFlags, dst2srcTcpFlags;
      u_int16_t src2dstFirstWin, src2dstLastWin, dst2srcFirstWin, dst2srcLastWin;
      TcpWindowStats src2dstWin, dst2srcWin;
    } tcp;

    struct {
      u_int32_t src2dstIcmpFlags, dst2srcIcmpFlags;  /* ICMP bitmask */
      u_int8_t src2dstIcmpType, dst2srcIcmpType;     /* ICMP type */
      u_int8_t src2dstIcmpCode, dst2srcIcmpCode;     /* ICMP code */
    } icmp;
  } protoCounters;

  struct {
    u_int32_t src2dst, dst2src;
  } payloadHash;

  struct {
    u_int16_t ssl_version, server_cipher;
    ndpi_cipher_weakness server_unsafe_cipher;
    char *server_name, *client_req_name;
    char *ja4c_hash;
  } ssl;

  struct {
    char *hassh_client, *hassh_server;
  } ssh;

  natFields nat;

  char *bittorrent_hash, *host_server_name;

  FlowDirection lastPktDirection; /* Direction of the last flow packet */
  FlowDirection beginInitiator, terminationInitiator;
  u_int32_t flags;                /* bitmask (internal) */

  PluginInformation *plugin;

  u_int16_t numCustomFields; /* 0 ... MAX_NUM_CUSTOM_FLOW_FIELDS */
  struct customNetFlowField *customFields; /* Populated on collected flows only */
  u_int32_t ingressVRFIF;
  u_int8_t flow_direction;
  u_int8_t forwarding_status;

#ifdef HAVE_NDPI_PRO
  struct {
    u_int8_t already_mapped:1, src_invalid_process:1, dst_invalid_process:1 /* not found during learning */, _notused:5;
    u_int32_t src_pid, src_uid, src_father_pid, src_father_uid,
      dst_pid, dst_uid, dst_father_pid, dst_father_uid;
    char *src_uid_name, *src_process, *src_father_process, *src_process_package_name,
      *src_father_uid_name,
      *src_father_package_name, *src_process_cmdline, *src_process_container_id,
      *dst_uid_name, *dst_process, *dst_father_process, *dst_process_package_name,
      *dst_father_uid_name,
      *dst_father_package_name, *dst_process_cmdline, *dst_process_container_id;
    float nw_latency_msec;
  } process_mapping;
#endif

#ifdef RCS
  RCSPluginsBucket rcs;
#endif
} FlowHashExtendedBucket;

/* *************************************** */

typedef struct flowHashBucket {
  u_int8_t magic, dont_export;

  FlowHashMicroBucket core;
  FlowHashExtendedBucket *ext;
} FlowHashBucket;

#endif /* _BUCKET_H_ */
