/*
 *  Copyright (C) 2002-25 - ntop.org
 *
 *  http://www.ntop.org/
 *
 * This code is proprietary code subject to the terms and conditions
 * defined in LICENSE file which is part of this source code package.
 */

/* ********************** */

#define MAX_NUM_NETWORKS                   128
#define CONST_INVALIDNETMASK                -1

/* ********************** */

#define DEFAULT_ZMQ_TCP_KEEPALIVE            1  /* Keepalive ON */
#define DEFAULT_ZMQ_TCP_KEEPALIVE_IDLE       30 /* Keepalive after 30 seconds */
#define DEFAULT_ZMQ_TCP_KEEPALIVE_CNT        3  /* Keepalive send 3 probes */
#define DEFAULT_ZMQ_TCP_KEEPALIVE_INTVL      3  /* Keepalive probes sent every 3 seconds */

/* ********************************************** */

#ifdef __linux__
#include <sys/time.h>
#endif

#define MAX_VARLEN_STR_LEN    4096 /* Headers might be long */
#define MAX_VARLEN_QUEUE        10  /* Number of elements */

typedef struct {
  char *str;
  u_int32_t str_len, seq_id;
} varlen_string_elem;

typedef struct {
  char *str;
  u_int32_t str_len;

  /* Sequence reordering */
  varlen_string_elem partial[MAX_VARLEN_QUEUE];
} varlen_string;

struct rfc822_info {
  varlen_string from, to, cc, subject, message_id, reply_to, date;
  varlen_string email_header;
  u_int8_t begin_data_cmd, email_header_processed, email_header_full;
};

typedef struct {
  u_int8_t packet_ready, rx_direction;
  u_int32_t packet_hash;
  struct pcap_pkthdr h;
  u_char *p;
  int packet_if_idx /* -1 = unknown */;
} QueuedPacket;

#define DEFAULT_QUEUE_CAPACITY  1024

typedef struct {
  u_int16_t insert_idx, remove_idx;
  u_int32_t num_insert, num_remove;
  ConditionalVariable dequeue_condvar;
  void *queueSlots;
} ItemsQueue;

typedef struct {
  u_int32_t value;
#ifndef HAVE_BUILTIN_ATOMIC
  pthread_rwlock_t lock;
#endif
} atomic_u_int32_t;

/* ************************************* */

typedef struct {
  u_int8_t mac[6];
  u_int32_t asn;
  UT_hash_handle hh;
} mac2asn_hash;

typedef struct {
  int vlan;
  u_int32_t input_interface_idx;
  u_int32_t output_interface_idx;
  UT_hash_handle hh;
} vlan2ifidx_hash;

void freeRfc822Info(struct rfc822_info *info);
void processEmailHeader(struct rfc822_info *info);
void dumpRfc822Info(struct rfc822_info *info);

#ifdef WIN32
#define nprobe_sleep(a /* sec */) { waitForNextEvent(1000*a /* ms */); }
extern unsigned long waitForNextEvent(unsigned long ulDelay /* ms */);
extern void initWinsock32();
extern short isWinNT();
#define close(fd) closesocket(fd)
char* printAvailableInterfaces(char* name_or_index);
char* getInterfaceDescription(char* name_or_index, char* buf, u_int buf_len);
#else
int nprobe_sleep(int secs);
#endif

extern void traceEvent(const int eventTraceLevel, const char* file, const int line, const char * format, ...);
extern void daemonize(void);
extern const char* topic2str(MessageTopicType topic);
/*
#ifndef WIN32
extern int snprintf(char *string, size_t maxlen, const char *format, ...);
#endif
*/
extern u_int8_t ip2mask(IpAddress *addr, HostInfo *ip);
#if defined(HAVE_MAXMINDDB)
extern void readASs(char *path);
extern void readCities(char *path);
#endif
extern V9V10TemplateElementId ver9_templates[];
extern void printTemplateInfo(V9V10TemplateElementId *templates, u_char show_private_elements);
extern void dumpPluginHelp(void);
extern void dumpPluginFamilies(void);
extern void dumpPluginStats(u_int timeDifference);
extern void dumpPluginTemplates(void);
extern void enablePlugins(void);
extern void setupPlugins(void);
extern void initAS(void);
extern void flowFilePrintf(V9V10TemplateElementId **templateList,
			   PluginEntryPoint *pluginEntryPoint,
			   FlowHashBucket *theFlow,
			   FlowDirection direction);
extern u_int8_t getFlowDirection(FlowHashBucket *theFlow, FlowDirection direction);
extern u_int16_t getVLANsrc(FlowHashBucket *theFlow);
extern u_int16_t getVLANdst(FlowHashBucket *theFlow);
extern void printTEID(u_int teid, FlowHashBucket *theFlow,
		      V9V10TemplateElementId *theTemplateElement,
		      char *element_key,
		      ndpi_serializer *serializer,
		      FlowDirection direction,
		      PluginEntryPoint *pluginEntryPoint);
extern void flowBufferPrintf(V9V10TemplateElementId **templateList,
			     PluginEntryPoint *pluginEntryPoint,
			     FlowHashBucket *theFlow,
			     FlowDirection direction,
			     ndpi_serializer *serializer);
extern void sanitizeV4Template(char *str);
extern float toMs(struct timeval *t);
extern void fromMs(struct timeval *t, u_int32_t ms);
extern u_int64_t toUs(struct timeval *t);
extern u_int32_t msTimeDiff(struct timeval *end, struct timeval *begin);
extern u_int32_t usTimeDiff(struct timeval *end, struct timeval *begin);
extern float timevalDiff(struct timeval *end, struct timeval *begin);
extern double timevalUsDiff(struct timeval *end, struct timeval *begin);
extern unsigned int ntop_sleep(unsigned int secs);
extern FlowHashBucket* getListHead(FlowHashBucket **list);
extern void addToList(FlowHashBucket *bkt, FlowHashBucket **list);
extern void parseInterfaceAddressLists(char* _addresses);
extern void parseLocalAddressLists(char* _addresses);
extern u_int32_t str2addr(char *address);
extern char* etheraddr_string(const u_char *ep, char *buf);
extern void fixTemplateToIPFIX(void);
extern char* getStandardFieldId(u_int id);
extern u_int16_t ifIdx(FlowHashBucket *theFlow, int computeInputIfIdx);
extern u_int32_t _ip2AS(IpAddress *addr);
extern u_int32_t _getAS(IpAddress *addr, HostInfo *bkt);
extern void bitmask_set(u_int32_t n, bitmask_selector* p);
extern void bitmask_clr(u_int32_t n, bitmask_selector* p);
extern u_int8_t bitmask_isset(u_int32_t n, bitmask_selector* p);

extern void loadApplProtocols(void);
extern u_int16_t port2ApplProtocol(u_int8_t proto, u_int16_t port);

extern void copyInt8(u_int8_t ndb_dump, u_int elementIdx, u_int8_t t8, char *outBuffer, uint *outBufferBegin, uint *outBufferMax);
extern void copyInt16(u_int8_t ndb_dump, u_int elementIdx, u_int16_t _t16, char *outBuffer, uint *outBufferBegin, uint *outBufferMax);
extern void copyInt32Dual(u_int8_t ndb_dump, u_int elementIdx, u_int32_t _t32,  u_int32_t _t32_ndb, char *outBuffer, uint *outBufferBegin, uint *outBufferMax);
extern void copyInt32(u_int8_t ndb_dump, u_int elementIdx, u_int32_t _t32, char* outBuffer, uint* outBufferBegin, uint* outBufferMax);
extern void copyInt64(u_int8_t ndb_dump, u_int elementIdx, u_int64_t _t64, char *outBuffer, uint *outBufferBegin, uint *outBufferMax);
extern void copyLen(u_int8_t ndb_dump, u_int elementIdx, u_char *str, int strLen, char *outBuffer, uint *outBufferBegin, uint *outBufferMax);


extern u_int64_t _htonll(u_int64_t v);
extern u_int64_t _ntohll(u_int64_t v);

extern int32_t gmt2local(time_t t);
extern void decrementLastPacket(FlowHashBucket *bkt, FlowDirection flow_direction, u_int len);
extern void resetBucketStats(FlowHashBucket* bkt,
			     const struct pcap_pkthdr *h,
			     u_char *p,
			     u_int len, u_int ip_offset, FlowDirection direction,
			     u_char *payload, int payloadLen);
extern void maximize_socket_buffer(int sock_fd, int buf_type);
extern u_int16_t getFlowApplProtocol(FlowHashBucket *theFlow);

/* bitmask */
extern void reset_bitmask(bitmask_selector *selector);
extern int alloc_bitmask(u_int32_t tot_bits, bitmask_selector *selector);
extern void free_bitmask(bitmask_selector *selector);

/* nprobe.c */
extern void parseBlacklistNetworks(char* _addresses);
extern u_int16_t isBlacklistedIPv4Address(u_int32_t addr) ;

#ifndef min
#define min(a, b) ((a > b) ? b : a)
#endif

#ifndef max
#define max(a, b) ((a > b) ? a : b)
#endif

#ifdef __linux__
extern void setCpuAffinity(char *dev_name, char *cpuId);
#endif

extern int mkdir_p(char *path);
extern void dropPrivileges(void);
extern void dumpLogEvent(LogEventType event_type, LogEventSeverity severity, char *message);
extern char* CollectorAddress2Str(CollectorAddress *collector, char *buf, u_int buf_len);
extern u_int32_t getCollectorUDPSocketDrops();
extern u_int64_t to_msec(struct timeval *tv);
extern u_int64_t to_usec(struct timeval *tv);
extern u_int64_t to_ntp_ts(struct timeval *tv);
extern char *getDummySystemId(void);

extern char* getSystemId(void);
extern struct timeval* min_nonzero_timeval(struct timeval *a, struct timeval *b);
extern struct timeval* max_timeval(struct timeval *a, struct timeval *b);
extern char* format_tv(struct timeval *a, char *buf, u_int buf_len);
extern char* get_ts_format(void);
extern char* compactEmailList(char *l);
extern char* formatFileTimestamp(time_t epoch, char *buf, u_int buf_len);
extern char* flowDirection2char(FlowDirection direction);
#ifdef HAVE_ZMQ
extern int initZMQPublisher(void);
extern int initZMQCollector();
extern void pollZMQIncomingMessages();
extern void sendZMQ(char *str, u_int32_t str_len,
		    u_int32_t flow_idx, MessageTopicType topic, u_int8_t is_binary_data);
extern void flushZMQBuffer(u_int32_t flow_idx, u_int8_t do_lock);
extern void flushZMQBufferedPackets();
#endif
extern char* detab(char *str);
extern float timeval2ms(struct timeval *tv);

extern u_int16_t getNumCores(void);
#ifdef HAVE_PF_RING
extern int forwardPacket(int rx_device_id, char *p, int p_len);
#endif
extern void freeVarLenStr(varlen_string *str);
extern void appendString(varlen_string *str, u_int32_t seq_id,
			 char *to_add, u_int to_add_len, u_int8_t zap_chars,
			 u_int8_t zap_trailing_carriage_return);
extern void appendRawString(varlen_string *str, u_int32_t seq_id, char *to_add,
			    u_int to_add_len, u_int8_t zap_chars);
extern int isStringFull(varlen_string *str);
extern int isStringEmpty(varlen_string *str);
extern char* varlen2str(varlen_string *str);
extern void flushVarlenString(varlen_string *str);
extern void removeDoubleChars(char *str, u_char charToZap);

#ifndef HAVE_STRNSTR
extern char* strnstr(const char *s, const char *find, size_t slen);
#endif

extern void initQueue(ItemsQueue *q);
extern void deleteQueue(ItemsQueue *q);

/* ****************************************************** */

#ifdef WIN32
static ticks getticks(void) {
  struct timeval tv;
  gettimeofday (&tv, 0);

  return (((ticks)tv.tv_usec) + (((ticks)tv.tv_sec) * 1000000LL));
}

#else
static __inline__ ticks getticks(void) {
#if defined(__i386__)
  ticks x;

  __asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));
  return x;
#elif defined(__x86_64__)
  u_int32_t a, d;

  asm volatile("rdtsc" : "=a" (a), "=d" (d));
  return (((ticks)a) | (((ticks)d) << 32));

  /*
    __asm __volatile("rdtsc" : "=A" (x));
    return (x);
  */
#else
  struct timeval tv;
  gettimeofday (&tv, 0);

  return (((ticks)tv.tv_usec) + (((ticks)tv.tv_sec) * 1000000LL));
#endif
}
#endif

/* ****************************************************** */

//#define PROFILING

#if defined(PROFILING) && defined(__linux__)
#define PROFILING_DECLARE(n) ticks __profiling_section_start[n]; char *__profiling_section_label[n]; ticks __profiling_section_tot[n]; u_int64_t __profiling_section_times[n]
#define PROFILING_INIT() memset(__profiling_section_label, 0, sizeof(__profiling_section_label)); memset(__profiling_section_tot, 0, sizeof(__profiling_section_tot)); memset(__profiling_section_times, 0, sizeof(__profiling_section_times))
#define PROFILING_SECTION_ENTER(l, i) __profiling_section_start[i] = getticks(), __profiling_section_label[i] = l
#define PROFILING_SECTION_EXIT(i)  __profiling_section_tot[i] += getticks() - __profiling_section_start[i], __profiling_section_times[i]++
#define PROFILING_SECTION_VAL(i)   __profiling_section_tot[i]
#define PROFILING_SECTION_AVG(i)   (__profiling_section_tot[i] / __profiling_section_times[i])
#define PROFILING_SECTION_CNT(i)   __profiling_section_times[i]
#define PROFILING_SECTION_LABEL(i) __profiling_section_label[i]
#else
#define PROFILING_DECLARE(n)
#define PROFILING_INIT()
#define PROFILING_SECTION_ENTER(l, i)
#define PROFILING_SECTION_EXIT(i)
#define PROFILING_SECTION_VAL(i)
#define PROFILING_SECTION_AVG(i)
#define PROFILING_SECTION_CNT(i)
#define PROFILING_SECTION_LABEL(i)
#endif

/* ****************************************************** */

extern void      initAtomic(atomic_u_int32_t *a);
extern u_int32_t incAtomic(atomic_u_int32_t *a, u_int32_t value);
extern u_int32_t decAtomic(atomic_u_int32_t *a, u_int32_t value);
extern u_int32_t getAtomic(atomic_u_int32_t *a);

/* ****************************************************** */

extern int execute_command(char *command_path, char *path, u_int8_t async);
extern void dump_bad_packet(const struct pcap_pkthdr *h, const u_char *p);
extern int bindthread2core(pthread_t thread_id, int core_id);
extern char* escapeQuotes(char *in, char *out, u_int out_len);
extern u_int32_t getPrimaryHostIPv4(void);
extern void getPrimaryHostIPv6(struct in6_addr *addr);
extern char* escapeJSON(char *in, char *out, u_int out_len, u_int8_t json_mode);
extern u_int16_t readIfIdx(char *ifname);
#ifdef DEBUG
extern void printIfMismatchPackets(const u_char *expectedPacket, int expectedLength,
				   const u_char *currentPacket, int currentLength);
#endif
extern u_int16_t icmpTypeCode(u_int8_t icmpType, u_int8_t icmpCode);
extern u_int64_t getIntVal(u_char *value, int len);
extern int cleanupJsonString(char *str, int maxlen, int len);
extern int checkCustomFieldPrint(V9V10TemplateElementId *theTemplateElement,
				 FlowDirection direction,
				 FlowHashBucket *bkt, char *line_buffer,
				 uint line_buffer_len,
				 u_int8_t json_mode);
extern int checkCustomFieldExport(u_int elementIdx,
				  V9V10TemplateElementId *theTemplateElement,
				  FlowDirection direction,
				  FlowHashBucket *bkt,
				  char *outBuffer,
				  uint *outBufferBegin,
				  uint *outBufferMax);
extern char* getISO8601Timestamp(struct timeval *when, char *buf,
				 u_int buf_len, u_int8_t add_msec);

/* ****************************************************** */

extern u_int32_t getAS(IpAddress *addr, HostInfo *bkt);
extern u_int32_t getAdjBGPAS(HostInfo *bkt);

/* ****************************************************** */

extern void add_nbar_appid_mapping(u_int32_t app_id, char *app_name);
extern char* find_nbar_appid_mapping(u_int32_t app_id, char *buf, u_int buf_len);

extern mac2asn_hash* load_mac2asn_mapping(const char *path);
extern int map_mac2asn(mac2asn_hash *h, u_int8_t *mac, u_int32_t *asn);
extern void free_mac2asn_mapping(mac2asn_hash*);

extern vlan2ifidx_hash* load_vlan2ifidx_mapping(const char *path);
extern int map_vlan2ifidx(vlan2ifidx_hash *h, int vlanId,
			  u_int32_t *input_idx, u_int32_t *output_idx);
extern void free_vlan2ifidx_mapping(vlan2ifidx_hash*);

#if defined(HAVE_NDPI_PRO) && defined(__linux__)
extern void updateFlowProcessInfo(FlowHashBucket *bkt);
#endif

extern time_t getFileModificationTime(char *path, u_int *len);

extern int gainCapabilities();
extern int dropCapabilities();
extern int retainCapabilities();
extern int isFileInUse(char* path);

#ifdef HAVE_MAXMINDDB
extern void free_GeoIPRecord(GeoIPRecord *geo);
#endif
#ifndef WIN32
int get_sockaddr(const IpAddress * const addr, struct sockaddr ** const sa, ssize_t * const sa_len);
#endif
extern long httpGet(const char * const url,
		    const char * const username, const char * const password,
		    int timeout,
		    char * const resp, const u_int resp_len);
extern char* symHostName(FlowHashBucket *theFlow, u_int8_t src_host, char *buf, u_int8_t buf_len);
extern char* formatMacAsWrId(FlowHashBucket *theFlow, char *buf, u_int8_t buf_len,
			     u_int8_t json_mode, u_int8_t *ret_len);
extern const char* flowEndReason2str(FlowEndReason r);
extern char* url_sym2num(char *url, char *buf, u_int buf_len);
extern float getFlowRTT(FlowHashBucket *bkt);
extern void  updateFlowLabel(FlowHashBucket *theFlow, u_int8_t fill_src_host);

extern void loadFlowLabels(const char *path);
extern void freeFlowLabels();

extern void fill_prefix(ndpi_prefix_t *p, int f, struct in_addr *a, int b, int mb);
extern ndpi_patricia_node_t* add_to_ptree(ndpi_patricia_tree_t *tree, int family, struct in_addr *addr, int bits, void *data);
extern int remove_from_ptree(ndpi_patricia_tree_t *tree, int family, struct in_addr *addr, int bits);
extern void* ptree_match(ndpi_patricia_tree_t *tree, int family, struct in_addr* addr, int bits);
extern void loadFlowLabels(const char *path);
extern void freeFlowLabels();

/* ips.c */
extern void loadIPSModeConfigFile(char *json);
extern void termIPSMode();
extern char* marker2str(u_int16_t m, char *buf, u_int buf_len);
extern u_int16_t setFlowVerdict(FlowHashBucket *bkt, ndpi_protocol proto_id);
extern int is_empty_mac(u_char addr[]);
extern void dump_hex(const u_char *p, int len);
  
#ifdef HAVE_NETFILTER
extern int openNFQ(u_int queue_id);
extern void closeNFQ();
extern int processNFQPacket();
#endif

extern ndpi_ptree_t* loadIPv4IPv6Addresses(char* _addresses);
extern u_int16_t isLocalIPv4Address(u_int32_t addr /* network byte order */);
extern u_int16_t isLocalIPv6Address(struct in6_addr ipv6);
extern u_int16_t isLocalAddress(IpAddress *addr);
extern u_int8_t getFlowExportDirection(FlowHashBucket *theFlow, FlowDirection direction);
extern u_int8_t isNumber(char *text);
extern char* getnDPIInfo(FlowHashBucket *bkt);
extern char* getnDPIDomainInfo(FlowHashBucket *bkt);
extern u_int32_t getApplicationId(FlowHashBucket *theFlow);

extern void saveAAAToCache(FlowHashBucket *theFlow);

/* TLS */
extern int initTLSServer(char *cert_pem_path, char *key_pem_path);
extern int initTLSClient(u_int8_t id);

extern int parseSNMPIinterfaceFile(char *path);
u_int8_t fromHex(char *in, u_int in_len, char *out, u_int out_len);

u_int32_t /* network byte order */ getInterfaceIPv4Address(char *ifname);
char* readFileIntoString(char *path);
char* urlEncode(const char *url);

void loadBlacklists();
bool isIPInBlacklist(IpAddress *addr);
u_int32_t getUniqueSourceId(FlowHashBucket *theFlow);

u_char* iso8859_to_utf8(u_char *in, u_char *out, u_int out_len) ;

char* str2lower(char* _str, bool duplicate_string);

/* ****************************************************** */
